import React, {useMemo} from "react";
import {AutoColumnContextProvider, Column, ColumnKey, Columns} from "./Column";
import {SmartableProperties, useTable} from "./Smartable";
import {RowInstance, RowLoader} from "./Row";
import {useAsyncManager} from "./AsyncManager";
import {ColumnHeading} from "./Columns/ColumnHeading";
import {sortRows} from "./Sort";

/**
 * Smartable instance component properties.
 */
export interface InstanceProperties<CK extends ColumnKey> extends SmartableProperties<CK>
{
	/**
	 * Table columns.
	 */
	columns: Columns<CK>;
}

/**
 * Main component for a Smartable table.
 */
export function Instance<CK extends ColumnKey>({columns}: InstanceProperties<CK>)
{
	return (
		<table className={"smartable"}>
			<thead>
			<ColumnsHeadings columns={columns} />
			</thead>
			<tbody>
			<TableBody />
			</tbody>
		</table>
	);
}

/**
 * Columns headings of a Smartable table.
 */
export function ColumnsHeadings<CK extends ColumnKey>({columns}: {columns: Columns<CK>})
{
	return (
		<>
			<tr className={"headings"}>
				{ // Showing title of each column.
					Object.keys(columns).map((key) => (
						<AutoColumnContextProvider key={key} columnKey={key}>
							<ColumnHeading/>
						</AutoColumnContextProvider>
					))
				}
			</tr>
			<tr className={"filters"}>
				{ // Add columns filters, if there are some.
					(Object.entries(columns) as [CK, Column][]).map(([columnKey, column]) => (
						column.filter && (
							<AutoColumnContextProvider key={columnKey as string} columnKey={columnKey}>
								<td>{column.filter.element}</td>
							</AutoColumnContextProvider>
						)
					))
				}
			</tr>
		</>
	);
}

export function TableBody<CK extends ColumnKey>()
{
	// Get data from table.
	const {data, columns, columnsSortState, columnsFilterStates} = useTable<CK>();

	// Get current data state from the async table value.
	const {currentDataState} = useAsyncManager<CK>(data);

	// Memorize filtered rows.
	const filteredRows = useMemo(() => (
		currentDataState.rows?.filter((row) => (
			// Checking each row to keep only those which match the filters.
			(Object.entries(columnsFilterStates) as [CK, any][]).every(([columnKey, filterState]) => (
				// For each filter, keep the row if data match the current filter.
				columns[columnKey].filter.filter(row.cells[columnKey].data, filterState)
			))
		))
	), [currentDataState.rows, columnsFilterStates]);

	// Memorize sorted rows.
	const sortedRows = useMemo(() => (
		sortRows<CK>(filteredRows, columns, columnsSortState)
	), [filteredRows, columns, columnsSortState]);

	return (
		sortedRows ? (
			sortedRows.map((rowData, index) => (
				// Rendering each row from its definition.
				<RowInstance row={rowData} />
			))
		) : (
			<RowLoader />
		)
	);
}
